#include "star_packet.h"
#include <unistd.h>
#include <signal.h>
#include <iconv.h>
#include <time.h>

#define TMP_FILE    "/tmp/newstar"
#define CONFIG_FILE "/etc/newstar.conf"
#define LOG_FILE    "/var/log/newstar"

const size_t msg_len = 1024;

char ifname[1024] = ""; //网卡id
char username[1024] = ""; //用户名
char password[1024] = ""; //密码
char message[msg_len]; //服务器返回消息
char cmd[1024] = "";
uint8_t dest_mac[6]; //目标mac地址
uint16_t magic; 
uint16_t echo_key;
packet *net; //libpcap的packet类

/* 编码转换函数 */
int convert( const char *desc, const char *src, char *input,
             size_t ilen, char *output, size_t olen) {
    char **pin = &input;
    char **pout = &output;
    iconv_t cd = iconv_open(desc, src);
    if (cd == (iconv_t)-1)
        return -1;
    memset(output, 0, olen);
    if (iconv(cd, pin, &ilen, pout, &olen)) return -1;
    iconv_close(cd);
    return 0;
}

/* mac 地址打印函数 */
void print_mac(const char *msg, uint8_t *mac) {
    printf("%s: %02X %02X %02X %02X %02X %02X\n",
            msg, mac[0], mac[1], mac[2], mac[3], mac[4], mac[5], mac[6]);
}

/* 登出函数 */
/* 发送 EAPOL-Logoff packet */
void logoff(int signo) {
    printf("正在注销...\n");
    uint8_t *buf = new uint8_t[buf_len];
    net->send( (uint8_t *)packet_logoff(net->mac, dest_mac), 
               sizeof(eapol_packet));
    printf("成功注销:)\n");
    exit(0);
}

void parentQuit(int signo){
    printf("开始连接... 连接情况请运行 newstar log 进行查看。\n");
    exit(0);
}

/* 认证函数 返回1表示认证成功,返回0表示认证失败,返回-1表示服务器应答不正确 */
/* 若认证成功或失败,message会设置成服务器返回的信息 */
/* 若认证成功, 服务器 mac 和 magic, echo_key 分别被设置 */
int auth(packet &net, char *username, char *password,
         char *message, uint8_t *dest_mac,
         uint16_t &magic, uint16_t &echo_key) {
    uint8_t *buf = new uint8_t[buf_len];
    uint8_t id;
    uint8_t key_len;
    uint8_t key[256];
    uint16_t _message_len;
    char _message[msg_len];
 
    eapol_packet *p;

    /* 发送 EAPOL-Start packet */
    printf("查找服务器... ");
    net.send( (uint8_t *)packet_start(net.mac, dest_mac),
              sizeof(eapol_packet));
    /* 接受返回 packet */
    if (net.recv(buf, buf_len) < 0)
        return -1;
    p = (eapol_packet *)buf;

    if (p->msg_type != SF_REQUEST) {
        /* 服务器返回错误 */
        return -1;
    }
    /* 保存服务器的 mac 地址 */
    memcpy(dest_mac, p->source_mac, 6);
    print_mac("服务器mac地址: ", dest_mac);
    /* 保存 session id */
    id = p->id;

    /* 发送 EAPOL-Response username packet */
    printf("发送用户名... ");
    net.send((uint8_t *)packet_username(net.mac, dest_mac, id, username),
              sizeof(eapol_packet));
    /* 接受返回 packet */
    if (net.recv(buf, buf_len) < 0)
        return -1;
    p = (eapol_packet *)buf;
    if (p->msg_type != SF_REQUEST) {
        /* 服务器返回错误 */
        return -1;
    }
    printf("用户名OK\n");
    /* 保存 session id */
    id = p->id;
    /* 保存 key_len */
    key_len = p->body[0x00];
    /* 保存 key */
    memcpy(key, &(p->body[0x01]), key_len);

    /* 发送 EAPOL-Response password packet */
    printf("发送密码... ");
    net.send((uint8_t *)packet_password(net.mac, dest_mac, id, 
                                        username, password,
                                        key, key_len), 
             sizeof(eapol_packet));
    if (net.recv(buf, buf_len) < 0)
        return -1;
    p = (eapol_packet *)buf;
    if (p->msg_type == SF_SUCCESS) {
        /* 服务器返回 success */
        /* 保存 message */
        _message_len = ntohs(*((uint16_t *)(p->body + 0x03)));
        memcpy(_message, p->body + 0x05, _message_len);
        _message[_message_len] = 0;
        /* 编码转换 */
        convert("UTF-8", "GBK", _message, msg_len, message, msg_len);
        /* 保存 magic 和 echo_key */
        uint16_t info_len = ntohs(*((uint16_t *)(p->body + _message_len + 0x09)));
        uint16_t skip = info_len + _message_len + 0x17;
        crypt(p->body + skip, 0x0F);
//        magic = ntohs(*((uint16_t *)(p->body + skip + 0x04)));
        magic = 0x102A;
        echo_key = ntohs(*((uint16_t *)(p->body + skip + 0x08)));
        return 1;
    }
    else if (p->msg_type == SF_FAILURE) {
        /* 服务器返回 failure */
        _message_len = ntohs(*((uint16_t *)(p->body + 0x03)));
        memcpy(_message, p->body + 0x05, _message_len);
        _message[_message_len] = 0;
        /* 编码转换 */
        convert("UTF-8", "GBK", _message, msg_len, message, msg_len);
        return 0;
    }
    else {
        /* 服务器返回错误 */
        return -1;
    }
}

void help(){
    printf("newstar for Linux @ 2009-01-14 情人节专版\n\n");
    printf("用法: \n");
    printf("    newstar [ stop | config | log | clearlog | help ]\n");
    printf("        不加参数        停止所有现运行的newstar并重新连接\n");
    printf("        stop            停止所有现运行的newstar\n");
    printf("        config          修改配置文件\n");
    printf("        log [n]         显示日志文件(最后n行, 默认n=20)\n");
    printf("        clearlog        清除日志文件\n");
    printf("        help -h --help  显示此消息\n\n");
    printf("    配置文件            %s\n", CONFIG_FILE);
    printf("    日志文件            %s\n\n", LOG_FILE);
    printf("  Code by felicia(feliciacrazy@qq.com) based on supertar\n");
    printf("  More functions added by felix021(felix021@gmail.com)\n\n");
    exit(0);
}

void inputLine(char var[]){ //输入一行; 回车则不修改原内容
    char tmp[1024];
    fflush(stdin);
    fgets(tmp, 1000, stdin);
    for(int i = 0; tmp[i]; ++i){
        if(tmp[i] == '\n') tmp[i] = 0;
    }
    if(tmp[0] != 0){
        strcpy(var, tmp);
        fflush(stdin);
    }
}

//进行设置
void config(){
    char ifname[10] = "";
    char username[100] = "";
    char password[100] = "";
    char tmp[1024] = "";

    //读取配置信息
    FILE * fp = fopen(CONFIG_FILE, "r");
    if(fp != NULL){
        fscanf(fp, "%[^\n]", ifname);fgetc(fp);
        fscanf(fp, "%[^\n]", username);fgetc(fp);
        fscanf(fp, "%[^\n]", password);fgetc(fp);
        fclose(fp);
        fp = NULL;
    }

    //输入配置信息
    printf("请输入网卡名(当前设置为%s, 按回车保持不变): ", ifname);
    inputLine(ifname);
    printf("请输入用户名(当前设置为%s, 按回车保持不变): ", username);
    inputLine(username);
    printf("请输入密码(按回车保持当前设置不变): ", password);
    inputLine(password);

    //写入配置信息
    printf("信息收集完毕, 正在写入...\n");
    fp = fopen(CONFIG_FILE, "w");
    if(fp == NULL){
        printf("配置文件无法写入，请重试!");
        return;
    }
    fprintf(fp, "%s\n", ifname);
    fprintf(fp, "%s\n", username);
    fprintf(fp, "%s\n", password);
    fclose(fp);
    printf("写入成功!\n");
    sprintf(cmd, "chmod 600 %s", CONFIG_FILE);
    system(cmd); //设置权限
}

//停止所有newstar进程(此进程除外)
void stop(){ 
    pid_t pid = getpid(), pid1;
    sprintf(cmd, "ps aux | grep newstar | grep -v grep | awk '{print $2}' > %s", TMP_FILE);
    system(cmd);
    FILE *fp = fopen(TMP_FILE, "r");
    while(fscanf(fp, "%d", &pid1) != EOF){
        if(pid == pid1) continue;
        kill(pid1, SIGTERM);
        //printf("PID %d killed\n", pid1);
    }
    fclose(fp);
    sprintf(cmd, "rm -rf %s", TMP_FILE);
    system(cmd);
}

//如果不存在配置文件，先进行设置
void detectConfig(){
    FILE *fp = NULL;
    fp = fopen(CONFIG_FILE, "r");
    if(fp == NULL){
        printf("第一次运行，请先配置帐号信息\n");
        config();
    }else
        fclose(fp);
}

//读取配置
void readConfig(char ifname[], char username[], char password[]){
    //打开配置文件
    FILE * fp = fopen("/etc/newstar.conf", "r");
    if(fp == NULL){
        printf("错误: 找不到配置文件\n请运行 newstar config 进行配置\n");
        exit(0);
    }
    //读入配置
    fscanf(fp, "%[^\n]", ifname);fgetc(fp);
    fscanf(fp, "%[^\n]", username);fgetc(fp);
    fscanf(fp, "%[^\n]", password);fgetc(fp);
    fclose(fp);
}

//连接
void connect(char ifname[], char username[], char password[]){

    /* 收到表示结束的信号时 logoff */
    signal(SIGHUP, logoff);  
    signal(SIGINT, logoff);
    signal(SIGQUIT, logoff);  
    signal(SIGABRT, logoff);
    signal(SIGKILL, logoff);  
    signal(SIGTERM, logoff);
    signal(SIGSTOP, logoff);  
    signal(SIGTSTP, logoff);

    net = new packet(ifname);

    //不断发送包裹向服务器请求连接
    while (1) {
        memcpy(dest_mac, broad_mac, sizeof(broad_mac));
        int ret = auth(*net, username, password, message, dest_mac, magic, echo_key);
        if (ret == -1) {
            printf("未能正确获得服务器的响应\n");
        } else if (ret == 0) {
            printf("验证失败\n服务器消息: %s\n", message);
        }else {
            printf("登录成功\n服务器消息: %s\n", message);
            break;
        }
        printf("等待5s...\n");
        sleep(5);
    }

    printf("保持连接...\n");
    printf("PID = %d\n", (int)getpid());
    fflush(stdout);

    //保持连接,每10s发送一次连接保持的数据包
    while (1) {
        net->send((uint8_t *)packet_echo(net->mac, dest_mac, magic, echo_key), 45);
        sleep(10);
    }
}

void viewlog(int lines){
    FILE *fp = fopen(LOG_FILE, "r");
    if(fp == NULL){
        printf("日志文件为空。\n");
        return;
    }
    char line[1024];
    int lc = 0;
    while(fgets(line, 1000, fp) != NULL) ++lc;
    if(lines <= 0 || lines >= lc) lines = lc;
    fseek(fp, 0, SEEK_SET);
    printf("日志文件(%s)长 %d 行，", LOG_FILE, lc);
    while(lc > lines) {
        fgets(line, 1000, fp);
        lc--;
    }
    printf("最后 %d 行内容如下: \n", lines);
    printf("------------------------------------\n");
    while(true){
        fscanf(fp, "%1000[^\n]", line);
        if(feof(fp)) break;
        puts(line);
        fgetc(fp);
    }
    printf("------------------------------------\n");
    fclose(fp);
}

void clearlog(){
    sprintf(cmd, "rm -rf %s", LOG_FILE);
    system(cmd);
    printf("日志文件已经清除。\n");
}

int main(int argc, char *argv[]) {
    if(geteuid() != 0){
        printf("错误: 需要使用root身份运行!\n");
        printf("请尝试在命令前加上sudo，或执行以下操作:\n");
        printf("  $ sudo chown root:root newstar\n");
        printf("  $ sudo chmod 4755 newstar\n");
        printf("  $ sudo cp newstar /usr/bin/\n"); 
        printf("然后即可正常使用 newstar。使用 newstar help 查看详细用法。\n\n");
        return 0;
    }
    if(argc > 1){
        if(strcmp(argv[1], "stop") == 0){  //停止
            stop();
            return 0;
        }else if(strcmp(argv[1], "config") == 0){ //配置
            printf("\n\n配置newstar:\n");
            config();
            printf("可键入 newstar 以连接网络。\n\n");
            return 0;
        }else if(strcmp(argv[1], "log") == 0){ //查看日志文件
            int lines = 20;
            if(argc > 2) sscanf(argv[2], "%d", &lines);
            viewlog(lines);
            return 0;
        }else if(strcmp(argv[1], "clearlog") == 0){ //删除日志文件
            clearlog();
            return 0;
        }else if(strcmp(argv[1], "help") == 0 || 
                 strcmp(argv[1], "-h") ||
                 strcpy(argv[1], "--help")){ //帮助
            help();
            return 0;
        }
    }

    //先停止其他所有newstar进程
    stop();

    //检测是否有配置文件，没有的话就进行设置
    detectConfig();

    //创建子进程，杀死父进程，子进程进行连接
    signal(SIGTERM, parentQuit);
    pid_t pid; //创建子进程
    if((pid = fork()) == -1){
        printf("错误: 无法创建新进程\n");
        return 0;
    }else if (pid == 0){ //子进程
        kill(getppid(), SIGINT); //杀死父进程
        //重定向输入到文件(子进程的输出看不到)
        freopen(LOG_FILE, "a", stdout); 
        printf("time @ %d\n", (int)time(NULL));
    }else{ //父进程，等死
        while(1) sleep(10);
    }

    //读取配置
    readConfig(ifname, username, password);

    //开始连接
    connect(ifname, username, password);

    return 0;
}
